# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def validate(conf):
    if conf.no_incremental and (conf.max_level or conf.always_level):
        raise Exception(
            'no-incremental option is not compatible '
            'with backup level options')

    if conf.action == "restore" and not conf.restore_abs_path and \
            not conf.nova_inst_id and not conf.cinder_vol_id and \
            not conf.cindernative_vol_id:
        raise Exception("Please provide restore_abs_path")

    if conf.restore_abs_path and not conf.action == "restore":
        raise Exception('Restore abs path with {0} action'
                        .format(conf.action))

    if conf.storage == "ssh" and \
            not (conf.ssh_key and conf.ssh_username and conf.ssh_host):
        raise Exception("Please provide ssh_key, "
                        "ssh_username and ssh_host")



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import json
from oslo_log import log
import requests.exceptions
import time

from freezer.storage import base

LOG = log.getLogger(__name__)


class SwiftStorage(base.Storage):
    """
    :type client_manager: freezer.osclients.ClientManager
    """

    def __init__(self, client_manager, container, work_dir, max_segment_size,
                 skip_prepare=False):
        """
        :type client_manager: freezer.osclients.ClientManager
        :type container: str
        """
        self.client_manager = client_manager
        # The containers used by freezer to executed backups needs to have
        # freezer_ prefix in the name. If the user provider container doesn't
        # have the prefix, it is automatically added also to the container
        # segments name. This is done to quickly identify the containers
        # that contain freezer generated backups
        if not container.startswith('freezer_'):
            self.container = 'freezer_{0}'.format(container)
        else:
            self.container = container
        self.segments = u'{0}_segments'.format(container)
        self.max_segment_size = max_segment_size
        super(SwiftStorage, self).__init__(work_dir, skip_prepare)

    def swift(self):
        """
        :rtype: swiftclient.Connection
        :return:
        """
        return self.client_manager.get_swift()

    def upload_chunk(self, content, path):
        """
        """
        # If for some reason the swift client object is not available anymore
        # an exception is generated and a new client object is initialized/
        # If the exception happens for 10 consecutive times for a total of
        # 1 hour, then the program will exit with an Exception.

        count = 0
        success = False
        while not success:
            try:
                LOG.info(
                    'Uploading file chunk index: {0}'.format(path))
                self.swift().put_object(
                    self.segments, path, content,
                    content_type='application/octet-stream',
                    content_length=len(content))
                LOG.info('Data successfully uploaded!')
                success = True
            except Exception as error:
                LOG.info(
                    'Retrying to upload file chunk index: {0}'.format(
                        path))
                time.sleep(60)
                self.client_manager.create_swift()
                count += 1
                if count == 10:
                    LOG.critical('Error: add_object: {0}'
                                 .format(error))
                    raise Exception("cannot add object to storage")

    def upload_manifest(self, backup):
        """
        Upload Manifest to manage segments in Swift

        :param backup: Backup
        :type backup: freezer.storage.base.Backup
        """
        self.client_manager.create_swift()
        headers = {'x-object-manifest':
                   u'{0}/{1}'.format(self.segments, backup)}
        LOG.info('Uploading Swift Manifest: {0}'.format(backup))
        self.swift().put_object(container=self.container, obj=str(backup),
                                contents=u'', headers=headers)
        LOG.info('Manifest successfully uploaded!')

    def upload_meta_file(self, backup, meta_file):
        # Upload swift manifest for segments
        # Request a new auth client in case the current token
        # is expired before uploading tar meta data or the swift manifest
        self.client_manager.create_swift()

        # Upload tar incremental meta data file and remove it
        LOG.info('Uploading tar meta data file: {0}'.format(
            backup.tar()))
        with open(meta_file, 'r') as meta_fd:
            self.swift().put_object(
                self.container, backup.tar(), meta_fd)

    def prepare(self):
        """
        Check if the provided container is already available on Swift.
        The verification is done by exact matching between the provided
        container name and the whole list of container available for the swift
        account.
        """
        containers_list = [c['name'] for c in self.swift().get_account()[1]]
        if self.container not in containers_list:
            self.swift().put_container(self.container)
        if self.segments not in containers_list:
            self.swift().put_container(self.segments)

    def info(self):
        ordered_container = {}
        containers = self.swift().get_account()[1]
        for container in containers:
            print(container)
            ordered_container['container_name'] = container['name']
            size = '{0}'.format((int(container['bytes']) / 1024) / 1024)
            if size == '0':
                size = '1'
            ordered_container['size'] = '{0}MB'.format(size)
            ordered_container['objects_count'] = container['count']
            print(json.dumps(
                ordered_container, indent=4,
                separators=(',', ': '), sort_keys=True))

    def meta_file_abs_path(self, backup):
        return backup.tar()

    def get_file(self, from_path, to_path):
        with open(to_path, 'ab') as obj_fd:
            iterator = self.swift().get_object(
                self.container, from_path,
                resp_chunk_size=self.max_segment_size)[1]
            for obj_chunk in iterator:
                obj_fd.write(obj_chunk)

    def remove(self, container, prefix):
        for segment in self.swift().get_container(container, prefix=prefix)[1]:
            self.swift().delete_object(container, segment['name'])

    def remove_backup(self, backup):
        """
            Removes backup, all increments, tar_meta and segments
            :param backup:
            :type backup: freezer.storage.base.Backup
            :return:
        """
        for i in range(backup.latest_update.level, -1, -1):
            if i in backup.increments:
                # remove segment
                self.remove(self.segments, backup.increments[i])
                # remove tar
                self.remove(self.container, backup.increments[i].tar())
                # remove manifest
                self.remove(self.container, backup.increments[i])

    def add_stream(self, stream, package_name, headers=None):
        i = 0
        for el in stream:
            self.upload_chunk(el, "{0}/{1}".format(package_name, "%08d" % i))
            i += 1
        if not headers:
            headers = {}
        headers['X-Object-Manifest'] = u'{0}/{1}/'.format(
            self.segments, package_name)
        headers['x-object-meta-length'] = len(stream)

        self.swift().put_object(self.container, package_name, "",
                                headers=headers)

    def find_all(self, hostname_backup_name):
        """
        :rtype: list[freezer.storage.base.Backup]
        :return: list of zero level backups
        """
        try:
            files = self.swift().get_container(self.container)[1]
            names = [x['name'] for x in files if 'name' in x]
            return [b for b in base.Backup.parse_backups(names, self)
                    if b.hostname_backup_name == hostname_backup_name]
        except Exception as error:
            raise Exception('Error: get_object_list: {0}'.format(error))

    def backup_blocks(self, backup):
        """

        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        try:
            chunks = self.swift().get_object(
                self.container, str(backup),
                resp_chunk_size=self.max_segment_size)[1]
        except requests.exceptions.SSLError as e:
            LOG.warning(e)
            chunks = self.client_manager.create_swift().get_object(
                self.container, str(backup),
                resp_chunk_size=self.max_segment_size)[1]

        for chunk in chunks:
            yield chunk

    def write_backup(self, rich_queue, backup):
        """
        Upload object on the remote swift server
        :type rich_queue: freezer.streaming.RichQueue
        :type backup: freezer.storage.base.Backup
        """
        for block_index, message in enumerate(rich_queue.get_messages()):
            segment_package_name = u'{0}/{1}/{2}/{3}'.format(
                backup, backup.timestamp,
                self.max_segment_size, "%08d" % block_index)
            self.upload_chunk(message, segment_package_name)
        self.upload_manifest(backup)

    def download_freezer_meta_data(self, backup):
        return {}

    def upload_freezer_meta_data(self, backup, meta_dict):
        pass



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
(C) Copyright 2016 Hewlett Packard Enterprise Development Company LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import abc
import datetime
import os
import six
import sys
import time

from freezer.openstack import backup
from freezer.openstack import restore
from freezer.snapshot import snapshot
from freezer.utils.checksum import CheckSum
from freezer.utils import exec_cmd
from freezer.utils import utils

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Job(object):
    """
    :type storage: freezer.storage.base.Storage
    :type engine: freezer.engine.engine.BackupEngine
    """

    def __init__(self, conf_dict, storage):
        self.conf = conf_dict
        self.storage = storage
        self.engine = conf_dict.engine

    def execute(self):
        start_time = utils.DateTime.now()
        LOG.info('Job execution Started at: {0}'.format(start_time))
        retval = self.execute_method()
        end_time = utils.DateTime.now()
        LOG.info('Job execution Finished, at: {0}'.format(end_time))
        LOG.info('Job time Elapsed: {0}'.format(end_time - start_time))
        return retval

    @abc.abstractmethod
    def execute_method(self):
        pass


class InfoJob(Job):
    def execute_method(self):
        self.storage.info()


class BackupJob(Job):
    def execute_method(self):
        try:
            (out, err) = utils.create_subprocess('sync')
            if err:
                LOG.error('Error while sync exec: {0}'.format(err))
        except Exception as error:
            LOG.error('Error while sync exec: {0}'.format(error))
        if not self.conf.mode:
            raise ValueError("Empty mode")
        mod_name = 'freezer.mode.{0}.{1}'.format(
            self.conf.mode, self.conf.mode.capitalize() + 'Mode')
        app_mode = importutils.import_object(mod_name, self.conf)
        backup_instance = self.backup(app_mode)

        level = backup_instance.level if backup_instance else 0

        metadata = {
            'curr_backup_level': level,
            'fs_real_path': self.conf.path_to_backup,
            'vol_snap_path': self.conf.path_to_backup,
            'client_os': sys.platform,
            'client_version': self.conf.__version__,
            'time_stamp': self.conf.time_stamp,
        }
        fields = ['action',
                  'always_level',
                  'backup_media',
                  'backup_name',
                  'container',
                  'container_segments',
                  'dry_run',
                  'hostname',
                  'path_to_backup',
                  'max_level',
                  'mode',
                  'backup_name',
                  'time_stamp',
                  'log_file',
                  'storage',
                  'mode',
                  'os_auth_version',
                  'proxy',
                  'compression',
                  'ssh_key',
                  'ssh_username',
                  'ssh_host',
                  'ssh_port',
                  'consistency_checksum'
                  ]
        for field_name in fields:
            metadata[field_name] = self.conf.__dict__.get(field_name, '') or ''
        return metadata

    def backup(self, app_mode):
        """

        :type app_mode: freezer.mode.mode.Mode
        :return:
        """
        backup_media = self.conf.backup_media

        time_stamp = utils.DateTime.now().timestamp
        self.conf.time_stamp = time_stamp

        if backup_media == 'fs':
            app_mode.prepare()
            snapshot_taken = snapshot.snapshot_create(self.conf)
            if snapshot_taken:
                app_mode.release()
            try:
                filepath = '.'
                chdir_path = os.path.expanduser(
                    os.path.normpath(self.conf.path_to_backup.strip()))

                if not os.path.isdir(chdir_path):
                    filepath = os.path.basename(chdir_path)
                    chdir_path = os.path.dirname(chdir_path)
                os.chdir(chdir_path)

                # Checksum for Backup Consistency
                if self.conf.consistency_check:
                    ignorelinks = (self.conf.dereference_symlink == 'none' or
                                   self.conf.dereference_symlink == 'hard')
                    consistency_checksum = CheckSum(
                        filepath, ignorelinks=ignorelinks).compute()
                    LOG.info('Computed checksum for consistency {0}'.
                             format(consistency_checksum))
                    self.conf.consistency_checksum = consistency_checksum

                hostname_backup_name = self.conf.hostname_backup_name
                backup_instance = self.storage.create_backup(
                    hostname_backup_name,
                    self.conf.no_incremental,
                    self.conf.max_level,
                    self.conf.always_level,
                    self.conf.restart_always_level,
                    time_stamp=time_stamp)
                self.engine.backup(filepath, backup_instance)
                return backup_instance
            finally:
                # whether an error occurred or not, remove the snapshot anyway
                app_mode.release()
                if snapshot_taken:
                    snapshot.snapshot_remove(
                        self.conf, self.conf.shadow,
                        self.conf.windows_volume)

        backup_os = backup.BackupOs(self.conf.client_manager,
                                    self.conf.container,
                                    self.storage)

        if backup_media == 'nova':
            LOG.info('Executing nova backup')
            backup_os.backup_nova(self.conf.nova_inst_id)
        elif backup_media == 'cindernative':
            LOG.info('Executing cinder backup')
            backup_os.backup_cinder(self.conf.cindernative_vol_id)
        elif backup_media == 'cinder':
            LOG.info('Executing cinder snapshot')
            backup_os.backup_cinder_by_glance(self.conf.cinder_vol_id)
        else:
            raise Exception('unknown parameter backup_media %s' % backup_media)
        return None


class RestoreJob(Job):

    def execute_method(self):
        conf = self.conf
        LOG.info('Executing FS restore...')
        restore_timestamp = None

        restore_abs_path = conf.restore_abs_path
        if conf.restore_from_date:
            restore_timestamp = utils.date_to_timestamp(conf.restore_from_date)
        if conf.backup_media == 'fs':
            backup = self.storage.find_one(conf.hostname_backup_name,
                                           restore_timestamp)
            self.engine.restore(backup, restore_abs_path, conf.overwrite)

            try:
                if conf.consistency_checksum:
                    backup_checksum = conf.consistency_checksum
                    restore_checksum = CheckSum(restore_abs_path,
                                                ignorelinks=True)
                    if restore_checksum.compare(backup_checksum):
                        LOG.info('Consistency check success.')
                    else:
                        raise ConsistencyCheckException(
                            "Backup Consistency Check failed: backup checksum "
                            "({0}) and restore checksum ({1}) did not match.".
                            format(backup_checksum, restore_checksum.checksum))
            except OSError as e:
                raise ConsistencyCheckException(
                    "Backup Consistency Check failed: could not checksum file"
                    " {0} ({1})".format(e.filename, e.strerror))
            return {}

        res = restore.RestoreOs(conf.client_manager, conf.container)
        if conf.backup_media == 'nova':
            res.restore_nova(conf.nova_inst_id, restore_timestamp)
        elif conf.backup_media == 'cinder':
            res.restore_cinder_by_glance(conf.cinder_vol_id, restore_timestamp)
        elif conf.backup_media == 'cindernative':
            res.restore_cinder(conf.cindernative_vol_id, restore_timestamp)
        else:
            raise Exception("unknown backup type: %s" % conf.backup_media)
        return {}


class ConsistencyCheckException(Exception):
    pass


class AdminJob(Job):

    def execute_method(self):
        if self.conf.remove_from_date:
            timestamp = utils.date_to_timestamp(self.conf.remove_from_date)
        else:
            timestamp = datetime.datetime.now() - \
                datetime.timedelta(days=self.conf.remove_older_than)
            timestamp = int(time.mktime(timestamp.timetuple()))

        self.storage.remove_older_than(timestamp,
                                       self.conf.hostname_backup_name)
        return {}


class ExecJob(Job):

    def execute_method(self):
        LOG.info('exec job....')
        if self.conf.command:
            LOG.info('Executing exec job....')
            exec_cmd.execute(self.conf.command)
        else:
            LOG.warning(
                'No command info options were set. Exiting.')
        return {}



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six

from oslo_log import log

from freezer.storage import base
from freezer.utils import utils

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class FsLikeStorage(base.Storage):
    DEFAULT_CHUNK_SIZE = 10000000

    def __init__(self, storage_directory, work_dir,
                 chunk_size=DEFAULT_CHUNK_SIZE, skip_prepare=False):
        self.storage_directory = storage_directory
        self.chunk_size = chunk_size
        super(FsLikeStorage, self).__init__(work_dir,
                                            skip_prepare=skip_prepare)

    def backup_to_file_path(self, backup):
        """

        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        return utils.path_join(self._zero_backup_dir(backup), backup)

    def _zero_backup_dir(self, backup):
        """
        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        return utils.path_join(self.storage_directory,
                               backup.hostname_backup_name,
                               backup.full_backup.timestamp)

    def prepare(self):
        self.create_dirs(self.storage_directory)

    def info(self):
        pass

    def meta_file_abs_path(self, backup):
        zero_backup = self._zero_backup_dir(backup)
        return utils.path_join(zero_backup, backup.tar())

    def upload_meta_file(self, backup, meta_file):
        """

        :param backup:
        :type backup: freezer.storage.base.Backup
        :param meta_file:
        :return:
        """
        zero_backup = self._zero_backup_dir(backup)
        to_path = utils.path_join(zero_backup, backup.tar())
        self.put_file(meta_file, to_path)

    def find_all(self, hostname_backup_name):
        backups = []
        backup_dir = utils.path_join(self.storage_directory,
                                     hostname_backup_name)
        self.create_dirs(backup_dir)
        timestamps = self.listdir(backup_dir)
        for timestamp in timestamps:
            increments = \
                self.listdir(utils.path_join(backup_dir, timestamp))
            backups.extend(base.Backup.parse_backups(increments, self))
        return backups

    def remove_backup(self, backup):
        """
        :type backup: freezer.storage.base.Backup
        :return:
        """
        self.rmtree(self._zero_backup_dir(backup))

    def write_backup(self, rich_queue, backup):
        """
        Stores backup in storage
        :type rich_queue: freezer.streaming.RichQueue
        :type backup: freezer.storage.base.Backup
        """
        filename = self.backup_to_file_path(backup)
        if backup.level == 0:
            self.create_dirs(self._zero_backup_dir(backup))

        with self.open(filename, mode='wb') as b_file:
            for message in rich_queue.get_messages():
                b_file.write(message)

    def backup_blocks(self, backup):
        """

        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        filename = self.backup_to_file_path(backup)
        with self.open(filename, 'rb') as backup_file:
            while True:
                chunk = backup_file.read(self.chunk_size)
                if chunk == '':
                    break
                if len(chunk):
                    yield chunk

    @abc.abstractmethod
    def listdir(self, directory):
        pass

    @abc.abstractmethod
    def put_file(self, from_path, to_path):
        pass

    @abc.abstractmethod
    def create_dirs(self, path):
        pass

    @abc.abstractmethod
    def rmtree(self, path):
        pass

    @abc.abstractmethod
    def open(self, filename, mode):
        pass

    def download_freezer_meta_data(self, backup):
        return {}

    def upload_freezer_meta_data(self, backup, meta_dict):
        pass



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import os
import re
import six

from oslo_log import log

from freezer.utils import utils

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Storage(object):
    """
    Any freezer storage implementation should be inherited from this abstract
    class.
    """

    def __init__(self, work_dir, skip_prepare=False):
        self.work_dir = work_dir
        if not skip_prepare:
            self.prepare()

    def download_meta_file(self, backup):
        """
        Downloads meta_data to work_dir of previous backup.
        :type backup: freezer.storage.base.Backup
        :param backup: A backup or increment. Current backup is incremental,
        that means we should download tar_meta for detection new files and
        changes. If backup.tar_meta is false, raise Exception
        :return:
        """
        utils.create_dir(self.work_dir)
        if backup.level == 0:
            return utils.path_join(self.work_dir, backup.tar())
        meta_backup = backup.full_backup.increments[backup.level - 1]
        if not meta_backup.tar_meta:
            raise ValueError('Latest update have no tar_meta')
        to_path = utils.path_join(self.work_dir, meta_backup.tar())
        if os.path.exists(to_path):
            os.remove(to_path)
        meta_backup.storage.get_file(
            meta_backup.storage.meta_file_abs_path(meta_backup), to_path)
        return to_path

    @abc.abstractmethod
    def meta_file_abs_path(self, backup):
        pass

    @abc.abstractmethod
    def get_file(self, from_path, to_path):
        pass

    @abc.abstractmethod
    def upload_meta_file(self, backup, meta_file):
        """
        :param backup:
        :type backup: freezer.storage.base.Backup
        :param meta_file:
        :return:
        """
        pass

    @abc.abstractmethod
    def upload_freezer_meta_data(self, backup, meta_dict):
        pass

    @abc.abstractmethod
    def download_freezer_meta_data(self, backup):
        pass

    @abc.abstractmethod
    def backup_blocks(self, backup):
        """
        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        pass

    @abc.abstractmethod
    def write_backup(self, rich_queue, backup):
        """
        :param rich_queue:
        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        pass

    @abc.abstractmethod
    def prepare(self):
        """
        Creates directories, containers
        :return: nothing
        """
        pass

    def find_one(self, hostname_backup_name, recent_to_date=None):
        """
        :param hostname_backup_name:
        :type hostname_backup_name: str
        :param recent_to_date:
        :type recent_to_date: int
        :rtype: Backup
        :return:
        """
        backups = self.find_all(hostname_backup_name)
        if recent_to_date:
            backups = [b for b in backups
                       if b.timestamp <= recent_to_date]
        err_msg = 'No matching backup name "{0}" found'\
            .format(hostname_backup_name)
        if not backups:
            raise IndexError(err_msg)
        backup = max(backups, key=lambda b: b.timestamp)
        last_increments = backup.increments.values()
        if recent_to_date:
            last_increments = [x for x in last_increments
                               if x.timestamp <= recent_to_date]
        return max(last_increments, key=lambda x: x.timestamp)

    @abc.abstractmethod
    def find_all(self, hostname_backup_name):
        """
        Gets backups by backup_name and hostname
        :param hostname_backup_name:
        :type hostname_backup_name: str
        :rtype: list[freezer.storage.base.Backup]
        :return: List of matched backups
        """
        pass

    @abc.abstractmethod
    def remove_backup(self, backup):
        """
        :type backup: freezer.storage.base.Backup
        :param backup:
        :return:
        """
        pass

    def remove_older_than(self, remove_older_timestamp, hostname_backup_name):
        """
        Removes backups which are older than the specified timestamp
        :type remove_older_timestamp: int
        :type hostname_backup_name: str
        """
        backups = self.find_all(hostname_backup_name)
        backups = [b for b in backups
                   if b.latest_update.timestamp < remove_older_timestamp]
        for b in backups:
            b.storage.remove_backup(b)

    @abc.abstractmethod
    def info(self):
        pass

    def create_backup(self, hostname_backup_name, no_incremental,
                      max_level, always_level, restart_always_level,
                      time_stamp=None):
        backups = self.find_all(hostname_backup_name)
        prev_backup = self._find_previous_backup(
            backups, no_incremental, max_level, always_level,
            restart_always_level)
        time_stamp = time_stamp or utils.DateTime.now().timestamp
        if prev_backup and prev_backup.tar_meta:
            return Backup(
                self, hostname_backup_name, time_stamp,
                prev_backup.level + 1, prev_backup.full_backup)
        else:
            return Backup(
                self, hostname_backup_name, time_stamp)

    @staticmethod
    def _find_previous_backup(backups, no_incremental, max_level, always_level,
                              restart_always_level):
        """

        :param backups:
        :type backups: list[freezer.storage.base.Backup]
        :param no_incremental:
        :param max_level:
        :param always_level:
        :param restart_always_level:
        :rtype: freezer.storage.base.Backup
        :return:
        """
        if no_incremental or not backups:
            return None
        incremental_backup = max(backups, key=lambda x: x.timestamp)
        """:type : freezer.storage.base.Backup"""
        latest_update = incremental_backup.latest_update
        if max_level and max_level <= latest_update.level:
            return None
        elif always_level and latest_update.level >= always_level:
                latest_update = \
                    incremental_backup.increments[latest_update.level - 1]
        elif restart_always_level and utils.DateTime.now().timestamp > \
                latest_update.timestamp + restart_always_level * 86400:
            return None
        return latest_update


class Backup(object):
    """
    Internal freezer representation of backup.
    Includes:
        name (hostname_backup_name) of backup
        timestamp of backup (when it was executed)
        level of backup (freezer supports incremental backup)
            Completed full backup has level 0 and can be restored without any
            additional information.
            Levels 1, 2, ... means that our backup is incremental and contains
            only smart portion of information (that was actually changed
            since the last backup)
        tar_meta - boolean value, that is true when we have available meta
        information.
        Please check for additional information about tar_meta
            http://www.gnu.org/software/tar/manual/html_node/Incremental-Dumps.html
    """
    PATTERN = r'(.*)_(\d+)_(\d+?)$'

    def __init__(self, storage, hostname_backup_name, timestamp, level=0,
                 full_backup=None, tar_meta=False):
        """
        :type storage: freezer.storage.base.Storage
        :param hostname_backup_name: name (hostname_backup_name) of backup
        :type hostname_backup_name: str
        :param timestamp: timestamp of backup (when it was executed)
        :type timestamp: int
        :param full_backup: Previous full_backup - not increment
        :type full_backup: Backup
        :param level: current incremental level of backup
        :type level: int
        :param tar_meta: Is backup has or has not an attached meta
        tar file in storage. Default = False
        :type tar_meta: bool
        :return:
        """
        if level == 0 and full_backup:
            raise ValueError("Detected incremental backup without level")
        self.hostname_backup_name = hostname_backup_name
        self._timestamp = timestamp
        self.tar_meta = tar_meta
        self._increments = {0: self}
        self._latest_update = self
        self._level = level
        self.storage = storage
        if not full_backup:
            self._full_backup = self
        else:
            self._full_backup = full_backup

    @property
    def full_backup(self):
        return self._full_backup

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def level(self):
        return self._level

    @property
    def increments(self):
        return self._increments

    @property
    def latest_update(self):
        return self._latest_update

    def tar(self):
        return "tar_metadata_{0}".format(self)

    def metadata(self):
        return self.storage.download_freezer_meta_data(self)

    def add_increment(self, increment):
        """

        :param increment:
        :type increment: Backup
        :return:
        """
        if self.level != 0:
            raise ValueError("Can not add increment to increment")
        if increment.level == 0:
            raise ValueError("Can not add increment with level 0")
        if (increment.level not in self._increments or
                increment.timestamp >
                self._increments[increment.level].timestamp):
            self._increments[increment.level] = increment
        if self.latest_update.level <= increment.level:
            self._latest_update = increment

    def __repr__(self):
        return '_'.join([self.hostname_backup_name,
                         repr(self._timestamp), repr(self._level)])

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def parse_backups(names, storage):
        """
        :param names:
        :type names: list[str] - file names of backups.
        :type storage: freezer.storage.base.Storage
        File name should be something like that host_backup_timestamp_level
        :rtype: list[freezer.storage.base.Backup]
        :return: list of zero level backups
        """
        prefix = 'tar_metadata_'
        tar_names = set([x[len(prefix):]
                         for x in names if x.startswith(prefix)])
        backup_names = [x for x in names if not x.startswith(prefix)]
        backups = []
        """:type: list[freezer.storage.base.BackupRepr]"""
        for name in backup_names:
            try:
                backup = Backup._parse(name)
                backup.tar_meta = name in tar_names
                backups.append(backup)
            except Exception as e:
                LOG.exception(e)
                LOG.error("cannot parse backup name: {0}".format(name))
        backups.sort(
            key=lambda x: (x.hostname_backup_name, x.timestamp, x.level))
        zero_backups = []
        """:type: list[freezer.storage.base.Backup]"""
        last_backup = None

        """:type last_backup: freezer.storage.base.Backup"""
        for backup in backups:
            if backup.level == 0:
                last_backup = backup.backup(storage)
                zero_backups.append(last_backup)
            else:
                if last_backup:
                    last_backup.add_increment(backup.backup(storage,
                                                            last_backup))
                else:
                    LOG.error("Incremental backup without parent: {0}"
                              .format(backup))

        return zero_backups

    @staticmethod
    def _parse(value):
        """
        :param value: String representation of backup
        :type value: str
        :return:
        """
        match = re.search(Backup.PATTERN, value, re.I)
        if not match:
            raise ValueError("Cannot parse backup from string: " + value)
        return BackupRepr(match.group(1), int(match.group(2)),
                          int(match.group(3)))

    def __eq__(self, other):
        if self is other:
            return True
        return type(other) is type(self) and \
            self.hostname_backup_name == other.hostname_backup_name and \
            self._timestamp == other.timestamp and \
            self.tar_meta == other.tar_meta and \
            self._level == other.level and \
            len(self.increments) == len(other.increments)


class BackupRepr(object):
    """
    Intermediate for parsing purposes - it parsed backup name.
    Difference between Backup and BackupRepr - backupRepr can be parsed from
    str and doesn't require information about full_backup
    """
    def __init__(self, hostname_backup_name, timestamp, level, tar_meta=False):
        """

        :param hostname_backup_name:
        :type hostname_backup_name: str
        :param timestamp:
        :type timestamp: int
        :param level:
        :type level: int
        :param tar_meta:
        :type tar_meta: bool
        :return:
        """
        self.hostname_backup_name = hostname_backup_name
        self.timestamp = timestamp
        self.level = level
        self.tar_meta = tar_meta

    def backup(self, storage, full_backup=None):
        """

        :param storage:
        :type storage: freezer.storage.base.Storage
        :param full_backup: freezer.storage.base.Backup
        :return:
        """
        return Backup(storage, self.hostname_backup_name, self.timestamp,
                      level=self.level, full_backup=full_backup,
                      tar_meta=self.tar_meta)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import subprocess

import tempest.test

class BaseFreezerTest(tempest.test.BaseTestCase):

    credentials = ['primary']

    def __init__(self, *args, **kwargs):

        super(BaseFreezerTest, self).__init__(*args, **kwargs)


    def setUp(self):
        super(BaseFreezerTest, self).setUp()

        self.get_environ()

    def tearDown(self):

        super(BaseFreezerTest, self).tearDown()

    @classmethod
    def get_auth_url(cls):
        return cls.os_primary.auth_provider.auth_client.auth_url[:-len('/tokens')]

    @classmethod
    def setup_clients(cls):
        super(BaseFreezerTest, cls).setup_clients()
        cls.get_environ()

    @classmethod
    def get_environ(cls):
        os.environ['OS_PASSWORD'] = cls.os_primary.credentials.password
        os.environ['OS_USERNAME'] = cls.os_primary.credentials.username
        os.environ['OS_PROJECT_NAME'] = cls.os_primary.credentials.tenant_name
        os.environ['OS_TENANT_NAME'] = cls.os_primary.credentials.tenant_name

        # Allow developers to set OS_AUTH_URL when developing so that
        # Keystone may be on a host other than localhost.
        if not 'OS_AUTH_URL' in os.environ:
                os.environ['OS_AUTH_URL'] = cls.get_auth_url()

        # Mac OS X uses gtar located in /usr/local/bin
        os.environ['PATH'] = '/usr/local/bin:' + os.environ['PATH']

        return os.environ


    def run_subprocess(self, sub_process_args, fail_message):

        proc = subprocess.Popen(sub_process_args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=self.environ, shell=False)

        out, err = proc.communicate()

        self.assertEqual(0, proc.returncode,
                         fail_message + " Output: {0}. "
                                        "Error: {1}".format(out, err))

        self.assertEqual('', out,
                       fail_message + " Output: {0}. "
                                      "Error: {1}".format(out, err))

        self.assertEqual('', err,
                         fail_message + " Output: {0}. "
                                        "Error: {1}".format(out, err))



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from freezer.utils import validator
from freezer.utils import utils

class TestValidator(unittest.TestCase):
    def test_pass(self):
        bunch = utils.Bunch()
        validator.validate(bunch)

    def test_no_increment(self):
        bunch = utils.Bunch(no_incremental=True, max_level=10)
        self.assertRaises(Exception, validator.validate, bunch)

    def test_restore_without_path(self):
        bunch = utils.Bunch(action="restore")
        self.assertRaises(Exception, validator.validate, bunch)

    def test_restore_with_path(self):
        bunch = utils.Bunch(action="restore", restore_abs_path="/tmp")
        validator.validate(bunch)

    def test_backup_with_restore_path(self):
        bunch = utils.Bunch(action="backup", restore_abs_path="/tmp")
        self.assertRaises(Exception, validator.validate, bunch)

    def test_ssh(self):
        bunch = utils.Bunch(storage="ssh", ssh_key="key", ssh_username="name",
                            ssh_host="localhost")
        validator.validate(bunch)

    def test_ssh_raises(self):
        bunch = utils.Bunch(storage="ssh", ssh_key="key", ssh_username="name")
        self.assertRaises(Exception, validator.validate, bunch)
        bunch = utils.Bunch(storage="ssh", ssh_key="key", ssh_host="localhost")
        self.assertRaises(Exception, validator.validate, bunch)
        bunch = utils.Bunch(storage="ssh", ssh_username="name",
                            ssh_host="localhost")
        self.assertRaises(Exception, validator.validate, bunch)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from freezer.storage import base
from freezer.storage import local
import mock


class TestBackup(unittest.TestCase):
    def test_backup_parse(self):
        self.assertRaises(ValueError, base.Backup._parse, "asdfasdfasdf")
        backup = base.Backup._parse("test_name_host_1234_0")
        self.assertEqual(backup.level, 0)
        self.assertEqual(backup.timestamp, 1234)
        self.assertEqual(backup.hostname_backup_name, "test_name_host")

    def test_backup_creation(self):
        backup = base.Backup(None, "name", 1234)
        self.assertEqual(backup.hostname_backup_name, "name")
        self.assertEqual(backup.timestamp, 1234)
        self.assertEqual(backup.level, 0)
        self.assertEqual(backup.latest_update.level, 0)
        self.assertEqual(backup.latest_update.timestamp, 1234)
        self.assertEqual(backup.latest_update.hostname_backup_name, "name")
        self.assertEqual(len(backup.increments), 1)

    def test_backup_full_backup(self):
        ok = False
        try:
            base.Backup(None, "name", 1324, 0, "full_backup")
        except ValueError:
            ok = True
        if not ok:
            raise Exception("Should throw ValueError")

    def test_backup_increment(self):
        backup = base.Backup(None, "name", 1234)
        self.assertRaises(ValueError, backup.add_increment, backup)
        increment = base.Backup(None, "name", 4567, 1, backup)
        backup.add_increment(increment)
        self.assertEqual(len(backup.increments), 2)

    def test__find_previous_backup(self):
        backup = base.Backup(None, "name", 1234)
        b = base.Storage._find_previous_backup([backup], False, 2, False, 0)
        assert b == backup

    def test__find_previous_backup_with_max_level(self):
        backup = base.Backup(None, "name", 1234)
        i1 = base.Backup(None, "name", 1234, 1, backup)
        i2 = base.Backup(None, "name", 1234, 2, backup)
        backup.add_increment(i1)
        backup.add_increment(i2)
        b = base.Storage._find_previous_backup([backup], False, 2, False, 0)
        assert not b

    def test__find_previous_backup_with_max_level_not_reached(self):
        backup = base.Backup(None, "name", 1234)
        i1 = base.Backup(None, "name", 1234, 1, backup)
        backup.add_increment(i1)
        b = base.Storage._find_previous_backup([backup], False, 2, False, 0)
        assert b == i1

    def test__find_previous_backup_with_always_level_reached(self):
        backup = base.Backup(None, "name", 1234)
        i1 = base.Backup(None, "name", 1234, 1, backup)
        i2 = base.Backup(None, "name", 1234, 2, backup)
        backup.add_increment(i1)
        backup.add_increment(i2)
        b = base.Storage._find_previous_backup([backup], False, False, 2, 0)
        assert b == i1

    def test__find_previous_backup_with_always_level_reached_2(self):
        backup = base.Backup(None, "name", 1234)
        i1 = base.Backup(None, "name", 1234, 1, backup)
        i2 = base.Backup(None, "name", 1234, 2, backup)
        backup.add_increment(i1)
        backup.add_increment(i2)
        b = base.Storage._find_previous_backup([backup], False, False, 3, 0)
        assert b == i2

    def test_add_increment_raises(self):
        backup = base.Backup(None, "name", 1234, level=3)
        self.assertRaises(ValueError, backup.add_increment, None)

    def test_restore_latest_backup(self):
        t = local.LocalStorage("", None, skip_prepare=True)
        t.find_all = mock.Mock()
        last = base.Backup(t, "host_backup", 5000)
        t.find_all.return_value = [
            base.Backup(t, "host_backup", 1000),
            base.Backup(t, "host_backup", 2000),
            base.Backup(t, "host_backup", 3000),
            base.Backup(t, "host_backup", 4000),
            base.Backup(t, "host_backup_f", 1000),
            last
        ]
        assert t.find_one("host_backup") == last

    def test_find_latest_backup_respects_increments_timestamp(self):
        test_backup = base.Backup(None, "host_backup", 5500)
        increment = base.Backup(None, "host_backup", 6000, 1, test_backup)
        test_backup.add_increment(increment)
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        t.find_all.return_value = [
            test_backup,
            base.Backup(None, "host_backup", 2000),
            base.Backup(None, "host_backup", 3000),
            base.Backup(None, "host_backup", 4000),
            base.Backup(None, "host_backup_f", 1000),
            base.Backup(None, "host_backup", 5000),
        ]
        assert t.find_one("host_backup") == increment

    def test_restore_from_date(self):
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        backup_restore = base.Backup(None, "host_backup", 3000)
        t.find_all.return_value = [
            base.Backup(None, "host_backup", 1000),
            base.Backup(None, "host_backup", 2000),
            backup_restore,
            base.Backup(None, "host_backup", 4000),
            base.Backup(None, "host_backup_f", 1000),
            base.Backup(None, "host_backup", 5000),
        ]
        assert t.find_one("host_backup", 3234) == backup_restore

    def test_restore_from_date_increment(self):
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        test_backup = base.Backup(None, "host_backup", 1000)
        increment = base.Backup(None, "host_backup", 3200, 1, test_backup)
        test_backup.add_increment(increment)
        t.find_all.return_value = [
            test_backup,
            base.Backup(None, "host_backup", 4000),
            base.Backup(None, "host_backup_f", 1000),
            base.Backup(None, "host_backup", 5000),
        ]
        assert t.find_one("host_backup", 3234) == increment

    def test__get_backups_wrong_name(self):
        result = base.Backup.parse_backups(["hostname"], None)
        assert len(result) == 0
        result = base.Backup.parse_backups(["hostname_100_2"], None)
        assert len(result) == 0

    def test__get_backups_good_name(self):
        result = base.Backup.parse_backups(["host_backup_100_0"], None)
        assert len(result) == 1
        result = result[0]
        assert result.hostname_backup_name == "host_backup"
        assert result.timestamp == 100
        assert result.level == 0

    def test_remove_older_than(self):
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        r1 = base.Backup(t, "host_backup", 1000)
        r2 = base.Backup(t, "host_backup", 2000)
        t.find_all.return_value = [
            r1,
            r2,
            base.Backup(t, "host_backup", 3000),
            base.Backup(t, "host_backup", 4000),
            base.Backup(t, "host_backup", 5000),
        ]
        t.remove_backup = mock.Mock()
        t.remove_older_than(3000, "host_backup")
        t.remove_backup.assert_any_call(r1)
        t.remove_backup.assert_any_call(r2)
        assert t.remove_backup.call_count == 2

    def test_create_backup(self):
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        t.find_all.return_value = []
        t._find_previous_backup = mock.Mock()
        t._find_previous_backup.return_value = \
            base.Backup(None, "host_backup", 3000, tar_meta=True)
        t.create_backup("", True, 12, False, False)

    def test_restart_always_level(self):
        t = local.LocalStorage(None, None, skip_prepare=True)
        t.find_all = mock.Mock()
        t.find_all.return_value = []
        backup = base.Backup(None, "host_backup", 3000, tar_meta=True)
        t._find_previous_backup([backup], False, None, None, 10)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
# (c) Copyright 2016 Hewlett-Packard Enterprise Development Company, L.P
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from freezer.openstack import osclients
from freezer.storage import swift
from freezer.storage import base


class TestSwiftStorage(unittest.TestCase):

    def setUp(self):
        opts = osclients.OpenstackOpts.create_from_env().get_opts_dicts()
        self.storage = swift.SwiftStorage(
            osclients.OSClientManager(opts.pop('auth_url'),
                                      opts.pop('auth_method', 'password'),
                                      **opts
                                      ),
            "freezer_ops-aw1ops1-gerrit0001.aw1.hpcloud.net",
            "/tmp/",
            100, skip_prepare=True
        )

        self.files = [
            "tar_metadata_hostname_backup_1000_0",
            "hostname_backup_1000_0",
        ]

        self.increments = [
            "tar_metadata_hostname_backup_1000_0",
            "hostname_backup_1000_0",
            "tar_metadata_hostname_backup_2000_1",
            "hostname_backup_2000_1",
        ]

        self.cycles_increments = [
            "tar_metadata_hostname_backup_1000_0",
            "hostname_backup_1000_0",
            "tar_metadata_hostname_backup_2000_1",
            "hostname_backup_2000_1",
            "tar_metadata_hostname_backup_3000_0",
            "hostname_backup_3000_0",
            "tar_metadata_hostname_backup_4000_1",
            "hostname_backup_4000_1",
        ]

        self.backup = base.Backup(self.storage,
                                  "hostname_backup", 1000, tar_meta=True,)
        self.backup_2 = base.Backup(self.storage,
                                    "hostname_backup", 3000, tar_meta=True)
        self.increment = base.Backup(self.storage,
                                     "hostname_backup", 2000,
                                     full_backup=self.backup,
                                     level=1,
                                     tar_meta=True)
        self.increment_2 = base.Backup(self.storage,
                                       "hostname_backup", 4000,
                                       full_backup=self.backup_2,
                                       level=1,
                                       tar_meta=True)

    def test__get_backups(self):
        backups = base.Backup.parse_backups(self.files, self.storage)
        self.assertEqual(1, len(backups))
        backup = backups[0]
        self.assertEqual(self.backup, backup)

    def test__get_backups_with_tar_only(self):
        backups = base.Backup.parse_backups(
            ["tar_metadata_hostname_backup_1000_0"], self.storage)
        self.assertEqual(0, len(backups))

    def test__get_backups_without_tar(self):
        backups = base.Backup.parse_backups(["hostname_backup_1000_0"],
                                            self.storage)
        self.assertEqual(1, len(backups))
        self.backup.tar_meta = False
        backup = backups[0]
        self.assertEqual(self.backup, backup)

    def test__get_backups_increment(self):
        backups = base.Backup.parse_backups(self.increments, self.storage)
        self.assertEqual(1, len(backups))
        self.backup.add_increment(self.increment)
        backup = backups[0]
        self.assertEqual(self.backup, backup)

    def test__get_backups_increments(self):
        backups = base.Backup.parse_backups(self.cycles_increments,
                                            self.storage)
        self.assertEqual(2, len(backups))
        self.backup.add_increment(self.increment)
        self.backup_2.add_increment(self.increment_2)
        self.assertEqual(self.backup, backups[0])
        self.assertEqual(self.backup_2, backups[1])
